@@ -9,6 +9,7 @@ import (
 	"os"
 	"runtime"
 	"strconv"
+	"strings"
 
 	"tailscale.com/types/lazy"
 	"tailscale.com/util/lineiter"
@@ -30,6 +31,8 @@ const (
 	WDMyCloud = Distro("wdmycloud")
 	Unraid    = Distro("unraid")
 	Alpine    = Distro("alpine")
+	UBNT      = Distro("ubnt") // Ubiquiti Networks
+	JetKVM    = Distro("jetkvm")
 )
 
 var distro lazy.SyncValue[Distro]
@@ -75,6 +78,12 @@ func linuxDistro() Distro {
 	case have("/usr/local/bin/freenas-debug"):
 		// TrueNAS Scale runs on debian
 		return TrueNAS
+	case have("/usr/bin/ubnt-device-info"):
+		// UBNT runs on Debian-based systems. This MUST be checked before Debian.
+		//
+		// Currently supported product families:
+		// - UDM (UniFi Dream Machine, UDM-Pro)
+		return UBNT
 	case have("/etc/debian_version"):
 		return Debian
 	case have("/etc/arch-release"):
@@ -95,10 +104,20 @@ func linuxDistro() Distro {
 		return Unraid
 	case have("/etc/alpine-release"):
 		return Alpine
+	case runtime.GOARCH == "arm" && isDeviceModel("JetKVM"):
+		return JetKVM
 	}
 	return ""
 }
 
+func isDeviceModel(want string) bool {
+	if runtime.GOOS != "linux" {
+		return false
+	}
+	v, _ := os.ReadFile("/sys/firmware/devicetree/base/model")
+	return want == strings.Trim(string(v), "\x00\r\n\t ")
+}
+
 func freebsdDistro() Distro {
 	switch {
 	case have("/etc/pfSense-rc"):
